/*
	decomp.c
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmd.h"
#include "decomp.h"
#include "debug.h"
#include "mem.h"
#include "types.h"


l_byte vm_variables[256];
boolean vm_flags[256];
char vm_strings[24][40];
int agi_horizon;
int foreground_color, background_color;

boolean text_mode = false, status_line_on = false, accept_input = false;
boolean dialogue_open = false;
int min_line = 1, user_input_line = 24, status_line = 0;

int main (void)
{
	logic_vm_t vm;
	int i;

	MemInit ();

	// Initialise virtual machine
	vm.IP = 0;
	vm.condition = NULL;
	vm.CS = getchar ();
	vm.CS += (getchar () << 8) + 2;
	vm.code = MemAlloc (vm.CS);

	fprintf (stderr, "Reading in LOGIC.....");
	fread (vm.code, 1, vm.CS, stdin);
	fprintf (stderr, "ok! [codesize = %i]\n", vm.CS);

	// Start virtual machine
	run_virtual_machine (&vm);

	// Clean-up
	while (vm.condition) {
		condition_pop (&vm);
		if (vm.condition) {
			for (i=0; i <= vm.condition->depth; i++)
				decomp ("\t");
		}
		decomp ("}\n");
	}

	decomp_flush ();
	condition_flush (&vm);

	return 0;
}

// WARNING: These condition functions bypass memory manager for SPEED!
void condition_push (logic_vm_t *vm, int startIP, int length)
{
	condition_t *new;

	new = (condition_t *) malloc (sizeof (condition_t));

	if (!new) {
		DEBUG_MSG1 ("FATAL - couldn't malloc for condition\n");
		condition_flush (vm);
		exit (1);
	}

	new->startIP = startIP;
	new->length = length;

	if (vm->condition)
		new->depth = vm->condition->depth + 1;
	else
		new->depth = 0;

	new->next = vm->condition;
	vm->condition = new;
}

void condition_pop (logic_vm_t *vm)
{
	condition_t *old;

	// Empty stack?
	if (!vm->condition) {
		DEBUG_MSG1 ("FATAL - empty condition stack\n");
		exit (1);
	}

	old = vm->condition;
	vm->condition = vm->condition->next;
	free (old);
}

void condition_flush (logic_vm_t *vm)
{
	condition_t *old;

	while (vm->condition) {
		old = vm->condition;
		vm->condition = vm->condition->next;
		free (old);
	}
}


#define BUF_SIZE	2048
char decomp_buffer[BUF_SIZE];
int decomp_size = 0, decomp_nl = 0;

void decomp (const char *fmt, ...)
{
	va_list ap;
	char buf[201], ch;
	int i, len;

	va_start (ap, fmt);
	vsnprintf (buf, 200, fmt, ap);
	va_end (ap);

	// Scan through, dealing with '\b'
	len = strlen (buf);
	for (i=0; i < len; i++) {
		ch = buf[i];
		if (ch == '\b') {
			decomp_size--;
		} else {
			decomp_buffer[decomp_size] = ch;
			decomp_size++;
		}
		if (ch == '\n')
			decomp_nl++;
	}

	if (decomp_nl > 3)
		decomp_flush ();
}

void decomp_flush (void)
{
	if (decomp_size > 0)
		fwrite (decomp_buffer, 1, decomp_size, stdout);
	fflush (stdout);
	decomp_size = 0;
	decomp_nl = 0;
}

void run_virtual_machine (logic_vm_t *vm)
{
	l_byte op, num_args, arg, t[2];
	int i, len;
	const vm_command_t *comm;
	boolean test;

//	while (vm->IP < vm->CS) {
while (1) {
		// Flush decompilation buffer
		decomp_flush ();

		// Get next command
		op = vm->code[vm->IP++];

		// End of conditional construct?
		if (vm->condition) {
			if ((vm->condition->startIP + vm->condition->length)
								< vm->IP)
			{
				condition_pop (vm);
				if (vm->condition) {
				for (i=0; i <= vm->condition->depth; i++)
					decomp ("\t");
				}
				decomp ("}\n");
			}
		}

		// Tab in for conditional constructs
		if (vm->condition) {
			for (i=0; i <= vm->condition->depth; i++)
				decomp ("\t");
		}

		// Conditional start? ('if' construct)
		if (op == 0xFF) {
			decomp ("if (");
			vm->IP--;
			test = logical_and (vm);

			// Read next two bytes for jump-information
			t[0] = vm->code[vm->IP++];
			t[1] = vm->code[vm->IP++];
			len = (t[1] << 8) + t[0];

			// Push this new conditional structure onto stack
			condition_push (vm, vm->IP, len);

			decomp (") {\n");
			if (test)
				continue;
		}

		// Unconditional jump? ('else', 'goto', etc.)
		if (op == 0xFE) {
			t[0] = vm->code[vm->IP++];
			t[1] = vm->code[vm->IP++];
			len = (t[1] << 8) + t[0];
			if (t[1] & 0x80) {
				// Find two's complement
				t[0] = ~t[0];
				t[1] = ~t[1];
				len = (t[1] << 8) + t[0] + 1;
				len = -len;
			}

			decomp ("JMP %i;\n", len);

			// If used as an 'else', end an 'if' bracket
			if (vm->condition)
				condition_pop (vm);
			if (vm->condition) {
				for (i=0; i <= vm->condition->depth; i++)
					decomp ("\t");
			}
			decomp ("}\n");
			continue;
		}

		// Valid command?
		if (op >= total_action_commands) {
			DEBUG_MSG4 ("CRITICAL - Unrecognised op 0x%02X "
				"[IP=%i CS=%i]\n", op, vm->IP-1, vm->CS);
			break;
		}

		// Get arguments
		comm = &action_commands[op];
		num_args = comm->args;

		decomp ("%s (", comm->name);
		for (i=0; i < num_args; i++) {
			arg = vm->code[vm->IP++];
			decomp ("%i, ", arg);
		}

		if (num_args > 0)
			decomp ("\b\b);\n");
		else
			decomp (");\n");

		if (op == 0)		// return statement
			break;
	}

	return;
}

boolean logical_branch (logic_vm_t *vm)
{
	boolean test;

	decomp ("(");
	switch (vm->code[vm->IP]) {
		case 0xFF:
			test = logical_and (vm);
			break;
		case 0xFD:
			decomp ("\b");
			test = logical_not (vm);
			return test;
			break;
		case 0xFC:
			test = logical_or (vm);
			break;
		default:
			test = false;
	}
	decomp (")");

	return test;
}

boolean logical_not (logic_vm_t *vm)
{
	boolean test;

	decomp ("!");
	vm->IP++;
	test = evaluate_next_condition (vm);
	test = !test;

	return test;
}

boolean logical_and (logic_vm_t *vm)
{
	l_byte op;
	boolean test;

	// Evaluate conditions until I reach the next 0xFF
	vm->IP++;

	do {
		test = false;
		op = vm->code[vm->IP];
		if (op >= total_test_commands) {
			// Nested test command
			test = logical_branch (vm);
		} else {
			test = evaluate_next_condition (vm);
		}

		decomp (" && ");
	} while (vm->code[vm->IP] != 0xFF);
	vm->IP++;

	decomp ("\b\b\b\b");

	return true;
}

boolean logical_or (logic_vm_t *vm)
{
	l_byte op;
	boolean test;

	// Evaluate conditions until I reach the next 0xFC
	vm->IP++;

	do {
		test = false;
		op = vm->code[vm->IP];
		if (op >= total_test_commands) {
			// Nested test command
			test = logical_branch (vm);
		} else {
			test = evaluate_next_condition (vm);
		}

		decomp (" || ");
	} while (vm->code[vm->IP] != 0xFC);
	vm->IP++;

	decomp ("\b\b\b\b");

	return true;
}

boolean evaluate_next_condition (logic_vm_t *vm)
{
	l_byte op, num_args, arg;
	int i;
	const vm_command_t *comm;
	boolean test;

	// Read in next op
	op = vm->code[vm->IP++];

	// Valid command?
	if (op >= total_test_commands) {
		DEBUG_MSG4 ("CRITICAL - Unrecognised op 0x%02X "
			"[IP=%i CS=%i]\n", op, vm->IP-1, vm->CS);
		return false;
	}

	// Get arguments
	comm = &test_commands[op];
	num_args = comm->args;

	// Kludge for 'said' (variable # of args)
	if (op == 0x0E) {
		test = handle_said (vm);
		return test;
	}

	decomp ("%s (", comm->name);
	for (i=0; i < num_args; i++) {
		arg = vm->code[vm->IP++];
		decomp ("%i, ", arg);
	}

	if (num_args > 0)
		decomp ("\b\b");
	decomp (")");

	test = true;
	return test;
}

boolean handle_said (logic_vm_t *vm)
{
	l_byte t[2];
	int num_args, i, args[10];	// Hope there's less than 10 args

	decomp ("said (");
	num_args = vm->code[vm->IP++];
	for (i=0; i < num_args; i++) {
		t[0] = vm->code[vm->IP++];
		t[1] = vm->code[vm->IP++];
		args[i] = (t[1] << 8) + t[0];
#if 0
if (0) {
	char *s;
	s = text_get_word (args[i]);
	decomp ("\"%s\", ", s);
	MemFree (s);
}
#else
		decomp ("0x%04X, ", args[i]);
#endif
	}

	if (num_args > 0)
		decomp ("\b\b");
	decomp (")");

	return true;
}
