/*
	driver_x11.c
*/

#include "config.h"
#include <signal.h>
#include <stdio.h>

#ifndef NO_X11		/* *HUGE* conditional! */

#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "debug.h"
#include "driver_main.h"
#include "driver_x11.h"
#include "events.h"
#include "keyboard.h"
#include "types.h"
#include "version.h"


unsigned int x_ega_palette[16][3] = {
{0, 0, 0},
{0, 0, 43690},
{0, 43690, 0},
{0, 43690, 43690},
{43690, 0, 0},
{43690, 0, 43690},
{43690, 21845, 0},
{43690, 43690, 43690},
{21845, 21845, 21845},
{21845, 21845, 65535},
{21845, 65535, 21845},
{21845, 65535, 65535},
{65535, 21845, 21845},
{65535, 21845, 65535},
{65535, 65535, 21845},
{65535, 65535, 65535}
};

struct x_key_mapping_t {
	KeySym keysym;
	int scancode;
} x_key_mapping[] = {
	{XK_Escape, SCANCODE_ESCAPE},

	{XK_1, SCANCODE_1}, {XK_2, SCANCODE_2}, {XK_3, SCANCODE_3},
	{XK_4, SCANCODE_4}, {XK_5, SCANCODE_5}, {XK_6, SCANCODE_6},
	{XK_7, SCANCODE_7}, {XK_8, SCANCODE_8}, {XK_9, SCANCODE_9},
	{XK_0, SCANCODE_0},

#if 0
	{XK_KP_Subtract, SCANCODE_MINUS},	/* *not* keypad! */
#endif
	{XK_KP_Equal, SCANCODE_EQUAL},

	{XK_BackSpace, SCANCODE_BACKSPACE}, {XK_Tab, SCANCODE_TAB},

	{XK_q, SCANCODE_Q}, {XK_w, SCANCODE_W}, {XK_e, SCANCODE_E},
	{XK_r, SCANCODE_R}, {XK_t, SCANCODE_T}, {XK_y, SCANCODE_Y},
	{XK_u, SCANCODE_U}, {XK_i, SCANCODE_I}, {XK_o, SCANCODE_O},
	{XK_p, SCANCODE_P},

	{XK_bracketleft, SCANCODE_BRACKET_LEFT},
	{XK_bracketright, SCANCODE_BRACKET_RIGHT},
	{XK_Return, SCANCODE_ENTER}, {XK_Control_L, SCANCODE_LEFTCONTROL},

	{XK_a, SCANCODE_A}, {XK_s, SCANCODE_S}, {XK_d, SCANCODE_D},
	{XK_f, SCANCODE_F}, {XK_g, SCANCODE_G}, {XK_h, SCANCODE_H},
	{XK_j, SCANCODE_J}, {XK_k, SCANCODE_K}, {XK_l, SCANCODE_L},

	{XK_semicolon, SCANCODE_SEMICOLON},
	{XK_apostrophe, SCANCODE_APOSTROPHE}, {XK_grave, SCANCODE_GRAVE},
	{XK_Shift_L, SCANCODE_LEFTSHIFT},
	{XK_backslash, SCANCODE_BACKSLASH},

	{XK_z, SCANCODE_Z}, {XK_x, SCANCODE_X}, {XK_c, SCANCODE_C},
	{XK_v, SCANCODE_V}, {XK_b, SCANCODE_B}, {XK_n, SCANCODE_N},
	{XK_m, SCANCODE_M},

	{XK_comma, SCANCODE_COMMA}, {XK_period, SCANCODE_PERIOD},
	{XK_slash, SCANCODE_SLASH}, {XK_Shift_R, SCANCODE_RIGHTSHIFT},
	{XK_KP_Multiply, SCANCODE_KEYPADMULTIPLY},
	{XK_Alt_L, SCANCODE_LEFTALT}, {XK_space, SCANCODE_SPACE},
	{XK_Caps_Lock, SCANCODE_CAPSLOCK},

	{XK_F1, SCANCODE_F1}, {XK_F2, SCANCODE_F2}, {XK_F3, SCANCODE_F3},
	{XK_F4, SCANCODE_F4}, {XK_F5, SCANCODE_F5}, {XK_F6, SCANCODE_F6},
	{XK_F7, SCANCODE_F7}, {XK_F8, SCANCODE_F8}, {XK_F9, SCANCODE_F9},
	{XK_F10, SCANCODE_F10},

	{XK_Num_Lock, SCANCODE_NUMLOCK},
	{XK_Scroll_Lock, SCANCODE_SCROLLLOCK},

	/* NumPad with NumLock off */
	{XK_KP_Home, SCANCODE_CURSORUPLEFT},
	{XK_KP_Up, SCANCODE_CURSORUP},
	{XK_KP_Page_Up, SCANCODE_CURSORUPRIGHT},
	{XK_KP_Left, SCANCODE_CURSORLEFT},
	{XK_KP_Right, SCANCODE_CURSORRIGHT},
	{XK_KP_End, SCANCODE_CURSORDOWNLEFT},
	{XK_KP_Down, SCANCODE_CURSORDOWN},
	{XK_KP_Page_Down, SCANCODE_CURSORDOWNRIGHT},

	/* NumPad with NumLock on */
	{XK_KP_7, SCANCODE_CURSORUPLEFT}, {XK_KP_8, SCANCODE_CURSORUP},
	{XK_KP_9, SCANCODE_CURSORUPRIGHT}, {XK_KP_4, SCANCODE_CURSORLEFT},
	{XK_KP_5, SCANCODE_KEYPAD5}, {XK_KP_6, SCANCODE_CURSORRIGHT},
	{XK_KP_1, SCANCODE_CURSORDOWNLEFT},
	{XK_KP_2, SCANCODE_CURSORDOWN},
	{XK_KP_3, SCANCODE_CURSORDOWNRIGHT}, {XK_KP_0, SCANCODE_KEYPAD0},

	{XK_F11, SCANCODE_F11}, {XK_F12, SCANCODE_F12},
	{XK_KP_Enter, SCANCODE_KEYPADENTER},
	{XK_Control_R, SCANCODE_RIGHTCONTROL},
	{XK_KP_Divide, SCANCODE_KEYPADDIVIDE},
	{XK_Sys_Req, SCANCODE_PRINTSCREEN}, {XK_Alt_R, SCANCODE_RIGHTALT},
	{XK_Pause, SCANCODE_BREAK},

	{0, 0}
};

Display *x_display;
Window x_root, x_window;
int x_screen;
Visual *x_visual;
Colormap x_colormap;
XSetWindowAttributes x_attributes;

long x_mapping[16];

int x11_init (void)
{
	int i;
	XColor xcol;
	Status err;

	XWMHints hints;
	XSizeHints sizehints;
	XClassHint classhint;
	XTextProperty appname, iconname;
	char *apptext = LAGII_TITLE " " LAGII_VERSION;
	char *icontext = LAGII_TITLE;

	x_display = XOpenDisplay (NULL);
	if (!x_display) {
		fprintf (stderr, "Can't open X display!\n");
		return -1;
	}

	x_root = DefaultRootWindow (x_display);
	x_screen = DefaultScreen (x_display);
	x_visual = DefaultVisual (x_display, x_screen);

	/* Check if visual is acceptable */
	if ((x_visual->class != PseudoColor) ||
	    (x_visual->bits_per_rgb > 8)) {
		DEBUG_MSG1 ("Need 8-bit PseudoColor visual!\n");
		XCloseDisplay (x_display);
		return -1;
	}

	/* Set EGA palette */
	x_colormap = XCreateColormap (x_display, x_root, x_visual,
							AllocNone);
	for (i=0; i < 16; i++) {
		/* Naughty! I assume XAllocColor always succeeds! */
		xcol.red = x_ega_palette[i][0];
		xcol.green = x_ega_palette[i][1];
		xcol.blue = x_ega_palette[i][2];
		err = XAllocColor (x_display, x_colormap, &xcol);
		if (!err)
			DEBUG_MSG2 ("Couldn't alloc color %i!\n", i);
	}

	/* Create window */
	x_attributes.event_mask = (KeyPressMask | KeyReleaseMask);
	x_window = XCreateWindow (x_display, x_root, 0, 0, 320, 200, 0,
			CopyFromParent, InputOutput, CopyFromParent,
			CWEventMask, &x_attributes);

	/* Set-up hints */
	XStringListToTextProperty (&apptext, 1, &appname);
	XStringListToTextProperty (&icontext, 1, &iconname);
	sizehints.flags = (PSize | PMinSize | PMaxSize);
	sizehints.min_width = 320;
	sizehints.max_width = 320;
	sizehints.min_height = 200;
	sizehints.max_height = 200;
	hints.flags = (StateHint | IconPixmapHint | InputHint);
	hints.initial_state = NormalState;
	hints.input = 1;
	classhint.res_name = "lagii";
	classhint.res_class = LAGII_TITLE;
	XSetWMProperties (x_display, x_window, &appname, &iconname,
			NULL, 0, &sizehints, &hints, &classhint);

	/* Map window */
	XMapRaised (x_display, x_window);

	/* Synchronise with X server */
	XSync (x_display, False);

	/* Just in case I get a nasty signal */
	signal (SIGHUP, x11_death);
	signal (SIGINT, x11_death);
	signal (SIGQUIT, x11_death);
	signal (SIGILL, x11_death);
	signal (SIGTERM, x11_death);

	return 0;
}

void x11_death (int sig)
{
	int i;
	struct sigmap_t {
		int signal;
		char *name;
	} sigmap[] = {
		{SIGHUP, "SIGHUP"},
		{SIGINT, "SIGINT"},
		{SIGQUIT, "SIGQUIT"},
		{SIGILL, "SIGILL"},
		{SIGTERM, "SIGTERM"},
		{0, "unknown"}
	};

	for (i=0;; i++) {
		if (sigmap[i].signal == 0)
			break;
		if (sigmap[i].signal == sig)
			break;
	}

	DEBUG_MSG3 ("Clobbered with signal %i (%s).\n", sig,
						sigmap[i].name);
	x11_close ();
	exit (-1);
}

void x11_close (void)
{
	XFreeColormap (x_display, x_colormap);
	XUnmapWindow (x_display, x_window);
	XDestroyWindow (x_display, x_window);
	XCloseDisplay (x_display);
}

void x11_blit (l_byte *buf)
{
	XImage *ximg;
	GC gc;
	int depth;

	gc = DefaultGC (x_display, x_screen);
	depth = DefaultDepth (x_display, x_screen);

	ximg = XCreateImage (x_display, x_visual, depth, ZPixmap, 0,
			(char *) buf, 320, 200, 8, 320);
	XPutImage (x_display, x_window, gc, ximg, 0, 0, 0, 0, 320, 200);
	XFree (ximg);
	XSync (x_display, False);
}

void x11_key_init (void)
{
	/* FIXME: Really do nothing? */
}

void x11_key_close (void)
{
	/* FIXME: Really do nothing? */
}

void x11_key_flush (void)
{
	/* FIXME: Really do nothing? */
}

void x11_key_update (void)
{
	XEvent ev;
	int i, scancode, press;
	KeySym key;
	char buf[2];

	while (XEventsQueued (x_display, QueuedAfterReading)) {
		XNextEvent (x_display, &ev);
		if ((ev.type != KeyPress) && (ev.type != KeyRelease))
			continue;
		if (ev.type == KeyPress)
			press = 1;
		else
			press = 0;
		XLookupString (&ev.xkey, buf, 1, &key, NULL);

		/* Kill shifted alpha keys */
		if ((key >= XK_A) && (key <= XK_Z))
			key += (XK_a - XK_A);

		/* Map this to a scancode */
		scancode = 0;
		for (i=0;; i++) {
			if (x_key_mapping[i].keysym == 0)
				break;
			if (x_key_mapping[i].keysym == key) {
				scancode = x_key_mapping[i].scancode;
				break;
			}
		}
		if (scancode == 0)
			continue;

		/* Transfer new key */
		event_key_handler (scancode, press);
	}
}


#endif	/* NO_X11 */
